// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2017, Javier Sánchez Pérez <jsanchez@ulpgc.es>
// All rights reserved.

#include<math.h>
#include<iostream>
#include<algorithm>
#include <random>

#include "midway_video_equalization.h"
#include "histogram.h"
#include "utils.h"
#include "gaussian/gaussian_conv_deriche.h"


/*****************************************************************************
 *                                                                           *
 *    Functions for Midway equalization through histogram inversion          *
 *                                                                           *
 *****************************************************************************/

/**
  *
  *  Function to calculate the contrast change functions
  * 
**/
float *compute_contrast_function(
  float *histograms, //accumulated histograms of the images
  float sigma,       //Gaussian standard deviation
  int   nframes      //number of frames
)
{
  float *f=new float[nframes*BIN_SIZE];
  int radius=2*sigma+0.5;
  
  //for each position in the 1st histogram we find the position in the 
  //other histograms and average the values using a Gaussian function
  #pragma omp parallel for
  for(int i=0;i<nframes;i++)
  {
    int hpos=i*BIN_SIZE;
    int min=(i-radius>0)?i-radius:0;
    int max=(i+radius<nframes)?i+radius:nframes-1;
    float *h1=&(histograms[hpos]);

    int *p2=new int[2*radius+1];
    for(int j=0;j<2*radius+1;j++)
      p2[j]=0;

    for(int p1=0;p1<BIN_SIZE;p1++)
    {
      //variables for the Gaussian convolution
      float sum=p1;
      float weight=1;
      
      //compute contrast change function with respect to previous frames
      for(int j=min;j<i;j++)
      {
        float *h2=&(histograms[j*BIN_SIZE]);
        int k=j-min;

        //find the corresponding level in histogram H(j)
        for(;p2[k]<BIN_SIZE && h2[p2[k]]<h1[p1]; p2[k]++);
        if(p2[k]==BIN_SIZE) 
           p2[k]--;
  
        //update the variables for the Gaussian convolution
        float norm=0.5*(i-j)*(i-j)/(sigma*sigma);
        float kernel=exp(-norm);
        weight+=kernel;
        sum+=kernel*p2[k];
      }

      //compute contrast change function with respect to following frames
      for(int j=i+1;j<=max;j++)
      {
        float *h2=&(histograms[j*BIN_SIZE]);
        int k=j-min;

        //find the corresponding level in histogram H(j)
        for(;p2[k]<BIN_SIZE && h2[p2[k]]<h1[p1]; p2[k]++);
        if(p2[k]==BIN_SIZE)
           p2[k]--;

        //update the variables for the Gaussian convolution
        float norm=0.5*(i-j)*(i-j)/(sigma*sigma);
        float kernel=exp(-norm);
        weight+=kernel;
        sum+=kernel*p2[k];
      }
      f[hpos+p1]=sum/weight;
    }
    delete []p2;  
  }
  
  //return the computed contrast change functions
  return f;
}


/**
  *
  *  Function to apply the ratio of gray values to the color image
  * 
**/
void gray2rgb(
  float gray,        //original pixel gray level 
  float smooth_gray, //smoothed pixel gray level
  unsigned char *I,  //pointer to the channels of the current pixel
  int nz             //number of channels
)
{
  if(gray<1) gray=1;
  if(smooth_gray<0) smooth_gray=0;

  //apply the contrast change to each color channel
  float *tmp=new float[nz];
  float max=0;

  //modify the color images according to the graylevel factor
  for(int c=0;c<nz;c++) {
    tmp[c]=(float)smooth_gray/gray*I[c];
    if(tmp[c]>max)
      max=tmp[c]; 
  }

  float factor=1;

  //if one channel goes beyond the pixel range, normalize them
  if(max>PIXEL_RANGE-1)
    factor=(PIXEL_RANGE-1)/max;
  
  for(int c=0;c<nz;c++) 
    I[c]=(unsigned char) ((float)factor*tmp[c]+0.5);

  delete []tmp;
}



/**
  *
  *  Function to apply the contrast change function to each image
  * 
**/
void apply_constrast_change(
  unsigned char *I, //input images
  float *Ig,        //grayscale images
  float *f,         //contrast change functions
  int nx,           //number of columns
  int ny,           //number of rows
  int nz,           //number of channels
  int nframes,      //number of frames
  int verbose       //verbose mode
)
{
  //compute the histogram of the rest of images
  #pragma omp parallel for
  for(int i=0;i<nframes;i++)
  {
    if(verbose && i%15==0) 
    {
      printf("  %d,",i);
      fflush(stdout);  
    }

    //get a reference to the actual frame
    unsigned char *J=&I[nx*ny*nz*i];
    
    //apply the contrast change to the pixels of the frame
    for(int j=0;j<nx*ny;j++)
    {
      int p1=nx*ny*i+j;
      int p2=j*nz;

      float prev=Ig[p1];

      //select relative position in the contrast change function
      int p=(int)(BIN_RATE*prev+0.5);
      if(p<0) p=0;
      else if (p>BIN_SIZE-1) p=BIN_SIZE-1;

      //get corresponding position in the contrast change function
      int q=i*BIN_SIZE+p;
      
      //obtain the corresponding contrast change in grayscale
      float post=(float)INV_BIN_RATE*f[q];
      
      //apply the contrast change to each color channel
      gray2rgb(prev, post, &J[p2], nz);

    }
  }  
}



/*****************************************************************************
 *                                                                           *
 *    Functions for Midway equalization through pixel sorting                *
 *                                                                           *
 *****************************************************************************/


/**
  *
  *  Function to sort the pixels of one image in ascending order
  *  It calculates the index of the sort array with respect 
  *  to the original positions in the image
  * 
**/
void sort_image(
  float  *I,      //input image
  size_t *index,  //output sorted indexes
  int    N        //number of pixels
) 
{
  //initialize index from 0 to N-1
  for(int i=0; i<N; i++)
    index[i]=i;

  //sort indexes based on I
  std::sort(
    index, &index[N],
    [I](size_t i1, size_t i2) {return I[i1]<I[i2];}
  );

  //randomize the index array to avoid strip artifacts
  int i=0;
  while(i<N-1)
  {
    int j=i+1;
    //select index subarray with the same intensity values
    while(j<N && I[index[i]]==I[index[j]]) j++;
    //randomize index subarray
    std::shuffle(&index[i],&index[j], std::mt19937());
    i=j;
  }

  //sort I based on the computed index
  float *Itmp=new float[N];
  for(int j=0;j<N;j++) 
    Itmp[j]=I[index[j]];

  //copy the temporal array into I
  for(int j=0;j<N;j++) 
    I[j]=Itmp[j];

  delete []Itmp;
}


/**
  *
  *  Function to sort the pixels of all the images 
  * 
**/
void sort_images(
  float  *I,      //input image
  size_t *index,  //output sorted indexes
  int    nframes, //number of frames
  int    N        //number of pixels
) 
{
  //sort every image and obtain index rank
  #pragma omp parallel for
  for(int i=0;i<nframes;i++)
    sort_image(&I[N*i], &index[N*i], N);
}


/**
  *
  *  Function to smooth the sorted images
  *  using the recursive Deriche's method
  * 
**/
void gaussian_convolution(
  float *I,      //input sorted images
  float *Is,     //output smoothed sorted images
  float sigma,   //Gaussian sigma
  int   nframes, //number of frames
  int   N        //number of pixels per image
)
{
  //This implementation is obtained from article:
  //  Pascal Getreuer, A Survey of Gaussian Convolution Algorithms, 
  //  Image Processing On Line, 3 (2013), pp. 286–310.
  //available at http://www.ipol.im/pub/art/2013/87/

  #pragma omp parallel for
  for(int i=0;i<N;i++)
  {
    deriche_coeffs c;
    num *buffer;
    
    //Gaussian convolution with Deriche's recursive filter
    deriche_precomp(&c, sigma, 3, 1E-10);
    buffer = new num[2 * nframes];
      
    //temporal convolution in each pixel separately
    deriche_gaussian_conv(c, &Is[i], buffer, &I[i], nframes, N);
    delete []buffer;  
  }
}


/**
  *
  *  Function to apply the contrast change to the images
  * 
**/
void rearrange_pixels(
  unsigned char *I, //input images
  float *Ig,        //sorted images
  float *Is,        //smoothed sorted images
  size_t *index,    //index of original images
  int nx,           //number of columns
  int ny,           //number of rows
  int nz,           //number of channels
  int nframes,      //number of frames
  int verbose       //verbose mode
)
{
  #pragma omp parallel for
  for(int i=0;i<nframes;i++)
  {
    if(verbose && i%15==0) 
    {
      printf("  %d,",i);
      fflush(stdout);  
    }

    //get a reference to the actual frame
    unsigned char *J=&I[nx*ny*nz*i];

    //apply the contrast change to the pixels of the frame
    for(int j=0;j<nx*ny;j++)
    {
      int p1=nx*ny*i+j;
      int p2=index[p1]*nz;

      //apply the contrast change to each color channel
      gray2rgb(Ig[p1], Is[p1], &J[p2], nz);
    }
  }  
}


/**
  *
  *  Function to control the three steps of equalization
  *  Calls both versions of equalization:
  *   - By histogram inversion
  *   - By sorting the pixels of the images
  * 
**/
void midway_video_equalization(
  unsigned char *I, //input images
  int nx,           //number of columns
  int ny,           //number of rows
  int nz,           //number of channels
  int nframes,      //number of frames
  int method,       //equalization method
  int sigma,        //Gaussian standard deviation
  int dithering,    //apply dithering 
  int verbose       //print messages
)
{ 
  int size=nx*ny*nframes;
  
  //allocate memory for grayscale images  
  float *Ig=new float[size];
  
  //convert images to graylevel
  rgb2gray(I, Ig, size);

  //apply dithering 
  if(dithering)
    add_noise(Ig, Ig, 2, size);
  
  if(method==HISTOGRAM_INVERSION_METHOD)
  {
    //allocate accumulated histograms for gray level images
    float *histograms=new float[BIN_SIZE*nframes];
    
    //compute accumulated histogram
    if(verbose) printf(" -Computing accumulated histograms\n");
    #pragma omp parallel for
    for(int i=0;i<nframes;i++)
      compute_normalized_histogram(
        &Ig[nx*ny*i], &(histograms[i*BIN_SIZE]),nx,ny
      );
    
    //compute contrast change functions
    if(verbose) printf(" -Computing contrast change functions\n");
    float *f=compute_contrast_function(histograms, sigma, nframes);

    //apply contrast change to the images
    if(verbose) printf(" -Applying contrast change functions:\n");
    apply_constrast_change(I, Ig, f, nx, ny, nz, nframes, verbose);
    
    delete []histograms;
    delete []f;
  }
  else 
  {
    int N=nx*ny;
    float *Is=new float[size];
    size_t *index=new size_t[size];
  
    //sort images and get indexes
    if(verbose) printf(" -Sorting the images\n");
    sort_images(Ig, index, nframes, N); 
    
    //compute contrast change functions
    if(verbose) printf(" -Temporal Gaussian convolution in each pixel\n");
    gaussian_convolution(Ig, Is, sigma, nframes, N);

    //rearrange pixels in their original positions
    if(verbose) printf(" -Rearranging pixels\n");
    rearrange_pixels(I, Ig, Is, index, nx, ny, nz, nframes, verbose);

    delete []Is;
    delete []index;    
  }
  
  delete []Ig;
}


